@@ -123,6 +123,8 @@ test_new (void)
 		TEST_EQ (class->kill_timeout, 5);
 		TEST_EQ (class->kill_signal, SIGTERM);
 
+		TEST_EQ (class->reload_signal, SIGHUP);
+
 		TEST_EQ (class->respawn, FALSE);
 		TEST_EQ (class->respawn_limit, 10);
 		TEST_EQ (class->respawn_interval, 5);
@@ -145,6 +147,8 @@ test_new (void)
 		TEST_EQ_P (class->setuid, NULL);
 		TEST_EQ_P (class->setgid, NULL);
 
+		TEST_EQ_P (class->apparmor_switch, NULL);
+
 		TEST_FALSE (class->deleted);
 
 		nih_free (class);
