@@ -32,7 +32,7 @@
 #include <nih/signal.h>
 #include <nih/main.h>
 #include "job.h"
-#include "test_util.h"
+#include "test_util_common.h"
 
 /*
  * To help with understanding the TEST_ALLOC_FAIL peculiarities
