@@ -39,6 +39,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <ctype.h>
+#include <time.h>
 
 #include <nih-dbus/dbus_error.h>
 #include <nih-dbus/dbus_connection.h>
@@ -57,6 +58,18 @@
 #error unable to find initctl binary as INITCTL_BINARY not defined
 #endif /* INITCTL_BINARY */
 
+static char *saved_xdg_config_home = NULL;
+static char *saved_xdg_runtime_dir = NULL;
+static char  test_xdg_config_home[PATH_MAX];
+static char  test_xdg_runtime_dir[PATH_MAX];
+
+/**
+ * test_setup_called:
+ *
+ * Set to TRUE if test_setup() called.
+ **/
+static int test_setup_called = FALSE;
+
 static void selfpipe_write (int n);
 static void selfpipe_setup (void);
 
@@ -119,6 +132,60 @@ wait_for_upstart (int session_init_pid)
 	TEST_EQ (running, TRUE);
 }
 
+/**
+ * session_init_reexec:
+ *
+ * @pid: pid of Session Init.
+ *
+ * Cause the Session Init running as pid @pid to re-exec.
+ **/
+void
+session_init_reexec (pid_t pid)
+{
+	nih_local NihDBusProxy *upstart = NULL;
+	char                   *address;
+	DBusConnection         *connection;
+	DBusMessage            *method_call;
+	DBusMessageIter         iter;
+	DBusError               error;
+
+	TEST_TRUE (pid);
+
+	TEST_TRUE (set_upstart_session (pid));
+	address = getenv ("UPSTART_SESSION");
+
+	TEST_TRUE (address);
+
+	connection = nih_dbus_connect (address, NULL);
+	TEST_NE_P (connection, NULL);
+
+	upstart = nih_dbus_proxy_new (NULL, connection,
+			NULL,
+			DBUS_PATH_UPSTART,
+			NULL, NULL);
+	TEST_NE_P (upstart, NULL);
+
+	method_call = dbus_message_new_method_call (upstart->name,
+			upstart->path,
+			"com.ubuntu.Upstart0_6", "Restart");
+	TEST_NE_P (method_call, NULL);
+
+	dbus_message_set_auto_start (method_call, upstart->auto_start);
+
+	dbus_message_iter_init_append (method_call, &iter);
+
+	/* Send the message, and wait for the reply. */
+	dbus_error_init (&error);
+
+	/* Don't bother checking reply - Upstart will sever the
+	 * connection anyway.
+	 */
+	(void)dbus_connection_send_with_reply_and_block (upstart->connection, method_call, -1, &error);
+
+	dbus_message_unref (method_call);
+	dbus_connection_unref (connection);
+}
+
 /* TRUE to denote that Upstart is running in user session mode
  * (FALSE to denote it's using the users D-Bus session bus).
  */
@@ -155,51 +222,51 @@ set_upstart_session (pid_t session_init_pid)
 	if (! getenv ("XDG_RUNTIME_DIR"))
 		return FALSE;
 
-	cmd = nih_sprintf (NULL, "%s list-sessions 2>&1", INITCTL_BINARY);
+	cmd = nih_sprintf (NULL, "%s list-sessions 2>&1", get_initctl_binary ());
 	TEST_NE_P (cmd, NULL);
 
 	/* We expect the list-sessions command to return a valid session
 	 * within a reasonable period of time.
 	 */
 	for (i = 0; i < loops; i++) {
-        sleep (1);
+		sleep (1);
 
 		RUN_COMMAND (NULL, cmd, &output, &lines);
 
-        if (lines < 1)
-            continue;
+		if (lines < 1)
+			continue;
 
-        /* Look for the specific session */
-        for (size_t line = 0; line < lines; lines++) {
+		/* Look for the specific session */
+		for (size_t line = 0; line < lines; lines++) {
 
-            /* No pid in output */
-            if (! isdigit(output[line][0]))
-                continue;
+			/* No pid in output */
+			if (! isdigit(output[line][0]))
+				continue;
 
-            pid = (pid_t)atoi(output[line]);
-            nih_assert (pid > 0);
+			pid = (pid_t)atoi(output[line]);
+			nih_assert (pid > 0);
 
-            if (pid != session_init_pid)
-                continue;
+			if (pid != session_init_pid)
+				continue;
 
-            /* look for separator between pid and value of
-             * UPSTART_SESSION.
-             */
-            value = strstr (output[0], " ");
-            if (! value)
-                continue;
+			/* look for separator between pid and value of
+			 * UPSTART_SESSION.
+			 */
+			value = strstr (output[0], " ");
+			if (! value)
+				continue;
 
-            /* jump over space */
-            value  += 1;
-            if (! value)
-                continue;
+			/* jump over space */
+			value  += 1;
+			if (! value)
+				continue;
 
-            /* No socket address */
-            if (strstr (value, "unix:abstract") == value) {
-                got = TRUE;
-                goto out;
-            }
-        }
+			/* No socket address */
+			if (strstr (value, "unix:abstract") == value) {
+				got = TRUE;
+				goto out;
+			}
+		}
 	}
 
 out:
@@ -268,6 +335,23 @@ selfpipe_setup (void)
     sigaction (SIGCHLD, &act, NULL);
 }
 
+/**
+ * have_timed_waitpid
+ *
+ * Return TRUE if precise timing information is available for timing
+ * tests.
+ **/
+int
+have_timed_waitpid (void)
+{
+	struct timespec res;
+
+	if (clock_getres (CLOCK_MONOTONIC_RAW, &res) < 0)
+		return FALSE;
+
+	return TRUE;
+}
+
 /**
  * timed_waitpid:
  *
@@ -349,9 +433,19 @@ get_initctl (void)
 {
 	static char path[PATH_MAX + 1024] = { 0 };
 	int         ret;
+	int         env_valid;
+
+	/* Sanity check calling environment */
+	if (test_user_mode) {
+		env_valid = getenv ("UPSTART_SESSION") ? TRUE : FALSE;
+	} else {
+		env_valid = getenv ("DBUS_SESSION_BUS_ADDRESS") ? TRUE : FALSE;
+	}
+
+	nih_assert (env_valid);
 
 	ret = sprintf (path, "%s %s",
-			INITCTL_BINARY,
+			get_initctl_binary (),
 			test_user_mode
 			? "--user"
 			: "--session");
@@ -384,7 +478,7 @@ _start_upstart (pid_t *pid, int user, char * const *args)
 	argv = NIH_MUST (nih_str_array_new (NULL));
 
 	NIH_MUST (nih_str_array_add (&argv, NULL, NULL,
-				UPSTART_BINARY));
+				get_upstart_binary ()));
 
 	if (args)
 		NIH_MUST (nih_str_array_append (&argv, NULL, NULL, args));
@@ -420,6 +514,7 @@ _start_upstart (pid_t *pid, int user, char * const *args)
  * @pid: PID of running instance,
  * @user: TRUE if upstart should run in User Session mode (FALSE to
  * use the users D-Bus session bus),
+ * @inherit_env: if TRUE, inherit parent environment,
  * @confdir: full path to configuration directory,
  * @logdir: full path to log directory,
  * @extra: optional extra arguments.
@@ -427,8 +522,9 @@ _start_upstart (pid_t *pid, int user, char * const *args)
  * Wrapper round _start_upstart() which specifies common options.
  **/
 void
-start_upstart_common (pid_t *pid, int user, const char *confdir,
-		      const char *logdir, char * const *extra)
+start_upstart_common (pid_t *pid, int user, int inherit_env,
+		      const char *confdir, const char *logdir,
+		      char * const *extra)
 {
 	nih_local char  **args = NULL;
 
@@ -446,14 +542,13 @@ start_upstart_common (pid_t *pid, int user, const char *confdir,
 					"--session"));
 	}
 
-	NIH_MUST (nih_str_array_add (&args, NULL, NULL,
-				"--no-startup-event"));
-
 	NIH_MUST (nih_str_array_add (&args, NULL, NULL,
 				"--no-sessions"));
 
-	NIH_MUST (nih_str_array_add (&args, NULL, NULL,
-				"--no-inherit-env"));
+	if (! inherit_env) {
+		NIH_MUST (nih_str_array_add (&args, NULL, NULL,
+					"--no-inherit-env"));
+	}
 
 	if (confdir) {
 		NIH_MUST (nih_str_array_add (&args, NULL, NULL,
@@ -486,7 +581,7 @@ start_upstart_common (pid_t *pid, int user, const char *confdir,
 void
 start_upstart (pid_t *pid)
 {
-	start_upstart_common (pid, FALSE, NULL, NULL, NULL);
+	start_upstart_common (pid, FALSE, FALSE, NULL, NULL, NULL);
 }
 
 /**
@@ -561,13 +656,21 @@ out:
 const char *
 get_upstart_binary (void)
 {
-	return UPSTART_BINARY;
+	static const char *upstart_binary = UPSTART_BINARY;
+
+	TEST_TRUE (file_exists (upstart_binary));
+
+	return upstart_binary;
 }
 
 const char *
 get_initctl_binary (void)
 {
-	return INITCTL_BINARY;
+	static const char *initctl_binary = INITCTL_BINARY;
+
+	TEST_TRUE (file_exists (initctl_binary));
+
+	return initctl_binary;
 }
 
 /**
@@ -740,3 +843,162 @@ search_and_replace (void        *parent,
 
 	return new;
 }
+
+/**
+ * file_exists:
+ * @path: file to check.
+ *
+ * Determine if specified file exists.
+ *
+ * Returns: TRUE if @path exists, else FALSE.
+ **/
+int
+file_exists (const char *path)
+{
+	struct stat  st;
+
+	nih_assert (path);
+
+	return ! stat (path, &st);
+}
+
+/**
+ * test_common_setup:
+ *
+ * Perform test setup.
+ *
+ * Currently, only needed for those tests which require unique xdg
+ * directories.
+ *
+ * If called, test must call test_common_cleanup() to cleanup.
+ **/
+void
+test_common_setup (void)
+{
+	char  *xdg_config_home;
+	char  *xdg_runtime_dir;
+
+	/* Take care to avoid disrupting users environment by saving and
+	 * restoring these variable (assuming the tests all pass...).
+	 */
+
+	xdg_config_home = getenv ("XDG_CONFIG_HOME");
+	if (xdg_config_home) {
+		nih_info ("Existing XDG_CONFIG_HOME found ('%s') - "
+				"saving for later restore",
+				xdg_config_home);
+
+		saved_xdg_config_home = NIH_MUST (nih_strdup (NULL, xdg_config_home));
+	}
+
+	TEST_FILENAME (test_xdg_config_home);
+	assert0 (mkdir (test_xdg_config_home, TEST_DIR_MODE));
+	assert0 (setenv ("XDG_CONFIG_HOME", test_xdg_config_home, 1));
+
+	nih_info ("Using test XDG_CONFIG_HOME='%s'", test_xdg_config_home);
+
+	xdg_runtime_dir = getenv ("XDG_RUNTIME_DIR");
+	if (xdg_runtime_dir) {
+		nih_info ("Existing XDG_RUNTIME_DIR found ('%s') - "
+				"saving for later restore",
+				xdg_runtime_dir);
+
+		saved_xdg_runtime_dir = NIH_MUST (nih_strdup (NULL, xdg_runtime_dir));
+	}
+
+	TEST_FILENAME (test_xdg_runtime_dir);
+	assert0 (mkdir (test_xdg_runtime_dir, TEST_DIR_MODE));
+	assert0 (setenv ("XDG_RUNTIME_DIR", test_xdg_runtime_dir, 1));
+
+	nih_info ("Using test XDG_RUNTIME_DIR='%s'", test_xdg_runtime_dir);
+
+	test_setup_called = TRUE;
+}
+
+/**
+ * test_common_cleanup:
+ *
+ * Perform cleanup of test setup.
+ *
+ * Currently, only needed for those tests which require unique xdg
+ * directories.
+ *
+ * If called, test must already have called test_common_setup().
+ **/
+void
+test_common_cleanup (void)
+{
+	nih_local char  *path = NULL;
+	char            *xdg_config_home;
+	char            *xdg_runtime_dir;
+	struct stat      statbuf;
+
+	if (! test_setup_called) {
+		nih_error ("BUG: Called %s without first calling %s",
+				__func__, "test_common_setup()");
+		abort ();
+	}
+
+	xdg_config_home = getenv ("XDG_CONFIG_HOME");
+	xdg_runtime_dir = getenv ("XDG_RUNTIME_DIR");
+
+	if (saved_xdg_config_home) {
+		nih_assert (test_xdg_config_home[0]);
+		TEST_EQ_STR (test_xdg_config_home, xdg_config_home);
+
+		if (stat (test_xdg_config_home, &statbuf)) {
+			nih_error ("A test has removed XDG_CONFIG_HOME '%s'", test_xdg_config_home);
+			abort ();
+		}
+
+		if (! S_ISDIR (statbuf.st_mode)) {
+			nih_error ("XDG_CONFIG_HOME '%s' no longer a directory", test_xdg_config_home);
+			abort ();
+		}
+
+		assert0 (rmdir (test_xdg_config_home));
+
+		assert0 (setenv ("XDG_CONFIG_HOME", saved_xdg_config_home, 1));
+		nih_info ("Restoring XDG_RUNTIME_DIR='%s'", saved_xdg_config_home);
+		nih_free (saved_xdg_config_home);
+		saved_xdg_config_home = NULL;
+	}
+
+	if (saved_xdg_runtime_dir) {
+		nih_assert (test_xdg_runtime_dir[0]);
+		TEST_EQ_STR (test_xdg_runtime_dir, xdg_runtime_dir);
+
+		if (stat (test_xdg_runtime_dir, &statbuf)) {
+			nih_error ("A test has removed XDG_RUNTIME_DIR '%s'", test_xdg_runtime_dir);
+			abort ();
+		}
+
+		if (! S_ISDIR (statbuf.st_mode)) {
+			nih_error ("XDG_RUNTIME_DIR '%s' no longer a directory", test_xdg_runtime_dir);
+			abort ();
+		}
+
+		path = NIH_MUST (nih_sprintf (NULL, "%s/upstart/sessions", xdg_runtime_dir));
+
+		if (! stat (path, &statbuf)) {
+			nih_local char *cmd = NULL;
+
+			/* Clean up if tests forgot to */
+			cmd = NIH_MUST (nih_sprintf (NULL, "rm %s/*.session 2>/dev/null", path));
+			system (cmd);
+
+			/* Remove the directory tree the first Session Init created */
+			assert0 (rmdir (path));
+			path = NIH_MUST (nih_sprintf (NULL, "%s/upstart", xdg_runtime_dir));
+			assert0 (rmdir (path));
+		}
+
+		assert0 (rmdir (test_xdg_runtime_dir));
+
+		assert0 (setenv ("XDG_RUNTIME_DIR", saved_xdg_runtime_dir, 1));
+		nih_info ("Restoring XDG_RUNTIME_DIR='%s'", saved_xdg_config_home);
+		nih_free (saved_xdg_runtime_dir);
+		saved_xdg_runtime_dir = NULL;
+
+	}
+}
